
//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     console_load.c
//
//  Purpose:    This module implements the LOAD command, which reads a
//              memory image into the current system.
//
//  History:    08/06/97    New header format standardization
//
//				06/09/99	Disallow if not in kernel mode.
//
//				06/11/99	Add BOOT command
//
//				07/13/99	Add RUN command
//

#include "vax.pch"
#include "console_proto.h"
#include "asmproto.h"
#include "memmap.h"
#include "imgdef.h"

#include <time.h>
void init_ihd_maps( void );

	
//	LOAD <filename>

long console_load( struct VAX ** Vax, char ** P )
{
    struct VAX * vax;
    struct VAX fvax;
    
    char * p, *fn;
    struct SYMBOL * sym;
    long size, n, rc, reading;
    long addr, pagecount;
    char buff[ 512 ];
    FILE * fp;
    
    vax = *Vax;
    p = *P;
    
    if( vax == 0L ) {
        return VAX_NOVAX;
    }


//	You can't do this if you aren't in kernel mode!  This is to prevent accidental
//	screw ups.

	if( vax-> pslw.cur_mod > 0 )
		return VAX_NOTKERNEL;

//  If there are unresolved forward references, disallow the
//  load operation.  The load could interfere with forward
//  references inappropriately.

    for( sym = vax-> console.symbols; sym; sym = sym-> next ) {
        if( sym-> forward != 0L ) 
            return VAX_ASMFORWARD;
    }
    
    
//  Do the work.  Start by fetching the filename.
    

    flush_blanks( &p );
    if( isend( vax, *p ))
        fn = "saved.vax";
    else {
        fn = p;
        for( fn = p; *p; p++ )
            if( *p == '\n' )
                *p = 0;
    }
    
    fp = fopen( fn, "rb" );
    if( fp == 0L ) {
        printf( "Can't open file \"%s\"\n", fn );
        return VAX_OK;
    }

//  Read the first eight bytes.  They must match a "magic number".

    size = fread( buff, 1, 8, fp );
    if( size != 8 || strncmp( buff, ";//EOF*\n", 8 ) != 0 ) {
        printf( "File \"%s\" is not a saved memory image.\n", fn );
        return VAX_OK;
    }

//  Read the version number.  Currently we only accept a version of 2.0

    size = fread( buff, 1, 2, fp );
    if( size != 2 )
        goto eof_exit;
        
    if( buff[ 0 ] != 2  || buff[ 1 ] != 0 ) {
        printf( "Unsupported file version %d.%d\n", buff[ 0 ], buff[ 1 ] );
        return VAX_OK;
    }

//  Read the copy of the VAX header at the time.

    size = fread( &fvax, sizeof( struct VAX ), 1, fp );
    if( size != 1 )
        goto eof_exit;

//  Blow away the vax we have, and make a new one.  This free's up memory,
//  cleans up the processor state, etc.

    printf( "Reading in a virtual VAX with %08lX (hex) bytes of memory.\n",
        fvax.memsize );
        
    free_vax( &vax );
    rc = alloc_vax( &vax, fvax.memsize );
    *Vax = vax;
    if( rc )
        return rc;

//  Copy the VM data

    for( n = 0; n < 3; n++ )
        vax-> region[ n ] = fvax.region[ n ];
    vax-> vm_initialized =  fvax.vm_initialized;

//  Copy the console data.  Empty out the pointers and
//  such that we don't want.

    sym = vax-> console.system_symbols;

    vax-> console = fvax.console;
    vax-> console.scale = 0L;
    vax-> console.breakpoint_list = 0L;
    vax-> console.prompt = 0L;
    vax-> console.CALL_active = 0;

    vax-> console.last_symbol = 0L;
    vax-> console.symbols = 0L;

    vax-> console.system_symbols = sym;

//  Copy the assembler flag data

    vax-> assembler.flags = fvax.assembler.flags;

//  Copy the register data.

    for( n = 0; n < MAXREG; n++ )
        vax-> reg[ n ] = fvax.reg[ n ];
    
    for( n = 0; n < MAXPRIVREG; n++ )
        vax-> preg[ n ] = fvax.preg[ n ];
    
    vax-> psl = fvax.psl;


//  Start reading page data

	reading = 1;
    while( reading ) {
    
        size = fread( &addr, sizeof( long ), 1, fp );
        if( size != 1 )
            goto eof_exit;
        
        size = fread( &pagecount, sizeof( long ), 1, fp );
        if( size != 1 )
            goto eof_exit;
        
        if( pagecount == 0L ) {
			reading = 0;
            break;
		}
        
        if( addr + pagecount * 512 - 1 > vax-> memsize ) {
            printf( "Saved memory image too large; error loading page at %08lX\n", addr );
			reading = 0;
            break;
        }
        
        size = fread( &( vax-> memory[ addr ]), 1, 512 * pagecount, fp );
        if( size != 512 )
            goto eof_exit;
        
        // printf( "Read %ld bytes at %08lx\n", pagecount * 512L, addr );
        
    
    }
        
    fclose( fp );

    return VAX_OK;

eof_exit:

    fclose( fp );
    printf( "Unexpected end-of-file reading \"%s\"\n", fn );
    return VAX_OK;
    
}


/*
 *	BOOT <filename>
 */

 
long console_boot( struct VAX ** Vax, char ** P )
{
    struct VAX * vax;
    
    char * p, *fn;

    long size, n, rc, reading;
    long addr, base;
	unsigned long flags;	/* Boot flags */
    unsigned char buff[ 512 ];
    FILE *  fp;
    
    vax = *Vax;
    p = *P;
    
    if( vax == 0L ) {
        return VAX_NOVAX;
    }


//	You can't do this if you aren't in kernel mode!  This is to prevent accidental
//	screw ups.

	if( vax-> pslw.cur_mod > 0 )
		return VAX_NOTKERNEL;

//	Zero out everything.

	rc = console_zero( Vax, P );
	if( rc )
		return rc;

	flush_blanks( &p );
	if( *p == '/' ) {
		p++;
		rc = asm_value( vax, &p, &flags, 0 );
		if( rc )
			return rc;
	}
	else
		flags = 0;


//  Do the work.  Start by fetching the filename.
    

    flush_blanks( &p );
    if( isend( vax, *p ))
        fn = "vmb.boot";
    else {
        fn = p;
        for( fn = p; *p; p++ )
            if( *p == '\n' )
                *p = 0;
    }
    
	*P = p;

    fp = fopen( fn, "rb" );
    if( fp == 0L ) {
        printf( "Can't open file \"%s\"\n", fn );
        return VAX_OK;
    }

//	Read in blocks of data.

	addr = base = 0x00001000L;

	reading = 1;
	rc = VAX_OK;
	while( reading ) {

		for( n = 0; n < 512; n++ )
			buff[ n ] = 0;

		size = fread( &(buff[ 0 ]), sizeof( unsigned char ), 512, fp );
	
		if( size <= 0 ) {
			reading = 0;
			for( n = 0; n< 512; n++ )
				if( !buff[ n ] )
					break;

#if 0
			printf( "n = %ld size = %ld ferror = %ld, feof = %ld\n", 
				n, size, ( long ) ferror( fp ), ( long ) feof( fp ));
#endif

			continue;
		}

		for( n = 0; n < size; n++ ) {
			vax-> memory[ addr++ ] = buff[ n ];
		}
		if( rc )
			break;
	}
        
    printf( "Bootstrap code %s loaded from address %08lX to %08lX\n", 
    	fn, base, addr - 1 );

    fclose( fp );

	vax-> R0 = 17;			/* UDA-50 boot device */
	vax-> R1 = 0;			/* TR number of adapter */
	vax-> R2 = 0x65001000;	/* Controller "A", unibus adapter ID 1000 */
	vax-> R3 = 0;			/* Boot device unit number */
	vax-> R4 = 0;			/* Boot LBN (0 for VMS) */
	vax-> R5 = flags;		/* Boot flags */
	vax-> R10 = 0;			/* Halt PC */
	vax-> R11 = 0;			/* Halt PSL */
	vax-> AP = 0;			/* Halt code */
	vax-> PC = base;		/* Base of VMB */
	vax-> SP = base;		/* Base of VMB */

	/* Turn off some stuff that the booter wouldn't expect or want */

	vax-> ICCS = 0;				/* No interval timer */
	vax-> clock_running = 0;	/* ditto */
	vax-> NICR = 0;				/* No next timer value */
	vax-> clock = 0;			/* No clock value */
	vax-> ICR = 0;				/* Ditto */
	vax-> TXCS = 0x00000080;	/* Console ready to xmit */
	vax-> SCBB = addr;			/* SCB defaults to after the VMB loader */

	vax-> IPL = 31;				/* Booter expects no interrupts at start */
	vax-> pslw.ipl = 31;
	
	/* Note that the 512 byte block before the SP and VMB load point is the */
	/* Restart Parameter Block (RPB).  This is zeroed in this case.  Need	*/
	/* find out later if this must be filled in some way.					*/

	/* We've loaded the booter, let 'er rip */
	
	printf( "Executing bootstrap code starting at %08lX\n", vax-> PC );
	
	execute_vax( vax, vax-> console.disasm );
	
	printf( "\n" );

    return rc;

}


/*
 *	RUN <filename>
 */

 
long console_run( struct VAX ** Vax, char ** P )
{
    struct VAX * vax;
    long debug;
    char * p, *fn;
	char sb[ 256 ], *sbp;
	
    long size, rc;
    long addr, paddr, naddr, base;
    long saved_mode, n, j, value, disp;
    long fixup_count;
	char nblocks;
	struct IHD ihd;
	long transfer[ 4 ];
	struct IHI ihi;
    struct ISD * isdlist = 0L;
    struct ISD * isd, static_isd;
 	struct ISD * fixup_isd;
 	struct IAF iaf;
 	char   shrname[ 40 ];
 	unsigned long file_addr;
 	long imageid = 0;
 	
 	struct SHR * shr, * shrnext;
 	struct SHR * shrlist = 0L;
 	
    FILE *  fp;
    short bsize = 512;
    
    vax = *Vax;
    p = *P;
    
    if( vax == 0L ) {
        return VAX_NOVAX;
    }
	debug = vax-> debug & DBG_IMAGES;
	
//	Initialize the structure mapping, if necessary

	init_ihd_maps();


//  Do the work.  Start by fetching the filename.
    

    flush_blanks( &p );
    if( isend( vax, *p ))
        fn = "vmb.boot";
    else {
        fn = p;
        if( *p == '"' )
        	p++;
        for( fn = p; *p; p++ )
            if( *p == '\n' || *p == '"' )
                *p = 0;
    }
    
	*P = p;

    fp = fopen( fn, "rb" );
    if( fp == 0L ) {
        printf( "Can't open file \"%s\"\n", fn );
        return VAX_OK;
    }

//	We have reserved the use of the scratch page for reading in image blocks.  Get
//	it's address now.

	strcpy( sb, "CONSOLE$SCRATCH" );
	sbp = sb;
	
	rc = get_symbol( vax, &sbp, &addr );
	if( rc ) {
		printf( "Unable to load image (no CONSOLE$SCRATCH area)\n" );
		return rc;
	}

//	Convert to physical address.  We'll need to be in Kernel mode for all this to
//	work correctly.
	
	saved_mode = vax-> pslw.cur_mod;
	set_mode_stack( vax, 0 );
	paddr = addr;
	rc = vm( vax, ( unsigned long * ) &paddr, &bsize, OP_MD );
	if( rc ) {
		printf( "Unable to resolve physical address of SYS$SCRATCH for I/O\n" );
		set_mode_stack( vax, saved_mode );
		return rc;
	}
	
//	Read in first header block.

	size = fread( &( vax-> memory[ paddr ]), 512L, 1, fp );
	if( size <= 0 ) {
		printf( "Unable to read image header\n" );
		return VAX_OK;
	}
	
//	Read the fixed portion of the header.  This is sort of ugly, but we have to do
//	it this way to get endian stuff right, etc.

	rc = load_memory( vax, addr + 16, ( void * ) &nblocks, 1 );
	if( rc ) goto abort;
	if( debug )
		printf( "There are %ld blocks in the header\n", nblocks );

//	Now that we know how big the goober is, let's reload it all at a better location.

	base = 0x00000200;
	fseek( fp, 0L, 0 );
	
	for( n = 0; n < nblocks; n++ ) {
	
		addr = base + ( n * 512 );
		paddr = addr;
		rc = vm( vax, ( unsigned long * ) &paddr, &bsize, OP_MD );
		if( rc ) {
			printf( "Unable to read entire header into memory\n" );
			goto abort;
		}
		
		size = fread( &( vax-> memory[ paddr ]), 512L, 1, fp );
		if( size <= 0 ) {
			printf( "Unable to read block %ld into memory at %08lX\n", n, addr );
			goto abort;
		}
		
	}
	
	if( debug )
		printf( "Read header into memory %08lX to %08lX\n", base, addr + 511 );
	
//	Map the fixed portion of the header.

	rc = map( vax, "IHD", ( void * ) &ihd, base );
	if( rc )
		goto abort;

//	Read in the transfer array data.

	for( n = 0; n < 4; n++ ) {
	
		rc = load_memory( vax, base + ihd.offset_transfer + ( n * 4 ), ( void * ) &(transfer[n]), 4 );
		if( rc )
			goto abort;
	}
	
	if( debug )
		printf( "Transfer array:\n\t%08lX\n\t%08lX\n\t%08lX\n\t%08lX\n",
			transfer[ 0 ], transfer[ 1 ], transfer[ 2 ], transfer[ 3 ]);

//	Get the image ident stuff

	addr = base + ihd.offset_ident;
	rc = map( vax, "IHI", ( void * ) &ihi, addr );
	if( rc )
		goto abort;
	
	ihi.imgnam[ ihi.imgnaml ] = 0;
	ihi.imgid [ ihi.imgidl  ] = 0;
	ihi.linkid[ ihi.linkidl ] = 0;
	
	if( debug )
		printf( "Image:\n\tid:   %s\n\tname: %s\n\tlinker: %s\n", 
			ihi.imgid, ihi.imgnam, ihi.linkid );
	
	
//	Walk the ISD list.

	addr = base + ihd.offset_ident + IHI_SIZE;

	if( debug )
		printf( "ISD list starts at %08lX\n", addr );

	for( n = 0; n < 1000; n++ ) {
	
		rc = map( vax, "ISD", ( void * ) &static_isd, addr );
		if( rc )
			goto abort;
			
		if( static_isd.size == 0 )
			break;
		addr = addr + static_isd.size;
		
		isd = ( struct ISD * ) getmem( sizeof( struct ISD ));
		*isd = static_isd;
		isd-> next = isdlist;
		isdlist = isd;
		
	}
	
	fixup_isd = 0L;
	
	for( isd = isdlist; isd; isd = isd-> next ) {
	
		if( debug ) {
			
			printf( "ISD:\n" );
			printf( "\tSize           %ld\n", ( long ) isd-> size );
			printf( "\tPages          %ld\n", ( long ) isd-> pages );
			addr = isd-> vpn << 9;
			
			printf( "\tVA             %08lX\n", ( long ) addr );
			printf( "\tPF cluster     %ld\n", ( long ) isd-> pfc );
			printf( "\tFlags          %08lX\n", ( long ) isd-> flags );
			
			if( isd-> flags & ISD_M_GBL )
				printf( "\t               GBL\n" );
			
			if( isd-> flags & ISD_M_CRF )
				printf( "\t               CRF\n" );
			
			if( isd-> flags & ISD_M_DZRO )
				printf( "\t               DZRO\n" );
			
			if( isd-> flags & ISD_M_WRT )
				printf( "\t               WRT\n" );
			
			if( isd-> flags & ISD_M_LASTCLU )
				printf( "\t               LASTCLU\n" );

			if( isd-> flags & ISD_M_COPYALWAY )
				printf( "\t               COPYALWAY\n" );
			
			if( isd-> flags & ISD_M_BASED )
				printf( "\t               BASED\n" );
			
			if( isd-> flags & ISD_M_FIXUPVEC )
				printf( "\t               FIXUPVEC\n" );
			
			if( isd-> flags & ISD_M_PROTECT )
				printf( "\t               PROTECT\n" );
			
			
			printf( "\tImage type     " );
			
			n = ISD_B_TYPE( isd-> flags );
			switch( n ) {
				case 0:		printf( "normal\n" );			break;
				case 1:		printf( "fixed shared\n" );		break;
				case 2:		printf( "fixed priv\n" );		break;
				case 3:		printf( "shared\n" );			break;
				case 4:		printf( "priv\n" );				break;
				case 253:
				case -3:  	printf( "user stack\n" );		break;
				default:    printf( "unknown type %d\n", n );		
			}
		
			if( n > 0 && n < 253 ) {
				isd-> name[ isd-> count ] = 0;
				printf( "\t\tsection name = %s\n", isd-> name );
			}
			
			printf( "\n" );

		}
		
		
//	Load this section if appropriate.

		if( isd-> flags & ISD_M_FIXUPVEC )
			fixup_isd = isd;

		n = ISD_B_TYPE( isd-> flags );		
		if( n != 0 )
			continue;
		
		file_addr = (isd-> vbn - 1) * 512;

		rc = fseek( fp, file_addr, SEEK_SET );
		if( rc ) {
			printf( "Unabled to seek to byte %08lX\n", file_addr );
		}
		
	    nblocks = isd-> pages;
	    
		for( n = 0; n < nblocks; n++ ) {
		
			addr = ( isd-> vpn << 9 ) + ( n * 512 );
			paddr = addr;
			rc = vm( vax, ( unsigned long * ) &paddr, &bsize, OP_MD );
			if( rc ) {
				printf( "Unable to read image section into memory\n" );
				goto abort;
			}
			
			size = fread( &( vax-> memory[ paddr ]), 1, 512, fp );
			if( size <= 0 ) {
				printf( "Unable to read block %ld into memory at %08lX\n", n, addr );
				goto abort;
			}
		}

	}
	
	fclose( fp );


//	Next, we'll need to work on the fixups....
	
	if( fixup_isd ) {
	
		addr = fixup_isd-> vpn << 9;
		rc = map( vax, "IAF", &iaf, addr );
		if( rc )
			goto abort;
		
		// Let's make the list of sharable images.
		
	
		naddr = addr + iaf.offset_shl;
		naddr = naddr + 16;  /* Hack, there are four longs I don't understand */
		rc = load_memory( vax, naddr, ( void * ) &iaf.offset_names, 4 );
		if( rc )
			goto abort;
			
		naddr = naddr + iaf.offset_names;
		
		/* Scoop up each name from the SHR structures */
		
		for( n = 1; n < iaf.shrimgcnt; n++ ) {
		
		    paddr = naddr + 0x08;  /* Two longs that seem to be always zero, maybe queue slots? */
		    naddr = naddr + 0x40;  /* Stride over each one */
		    
		    for( j = 0; j < 40; j++ ) {
			    rc = load_byte( vax, paddr + j, ( void * ) &shrname[ j ] );
		   	}
		   	
		    shrname[ shrname[ 0 ] + 1 ] = 0;
		    
		    shr = ( struct SHR * ) getmem( sizeof( struct SHR ));
		    shr-> next = shrlist;
		    shrlist = shr;
		    strcpy( shr-> name, &( shrname[ 1 ]));
		    imageid++;
		    shr-> id = imageid;
		    
		    if( debug )
		    	printf( "Sharable image %ld:  %s\n", shr-> id, shr-> name );
		}
		
		/*  Now we can run the G^ list and bind each one to a sharable image */
		
		naddr = addr + iaf.offset_g_fix;
		
		/* Get the number of entries in the next list */
		rc = load_memory( vax, naddr, ( void * ) &fixup_count, 4 );
		if( rc )
			goto abort;
		
		while( fixup_count ) {
			
			
			/* Get which image we're doing fixup's for */
			
			naddr += 4;
			rc = load_memory( vax, naddr, ( void * ) &imageid, 4 );
			if( rc )
				goto abort;
			
		    for( shr = shrlist; shr; shr = shr-> next )
		    	if( imageid == shr-> id )
		    		break;
			    
			
			/* Now loop over all the fixups */
			
			for( n = 0; n < fixup_count; n++ ) {

			    naddr += 4;
			    rc = load_memory( vax, naddr, ( void * ) &paddr, 4 );
			    if( rc )
			    	goto abort;
			    
			    			    
			    if( debug )
			    	printf( "Fixup at address %08lX to offset %08lX in image %s\n", 
			    		naddr, paddr, shr-> name );
			    
			    sprintf( sb, "SHIM$%s_%08lX", shr->name, paddr );
			    sbp = sb;
			    
			    disp = K_ADDR_L;
			    vax-> console.deposit = naddr;
			    rc = asm_value( vax, &sbp, ( unsigned long * )  &value, disp );
			    if( rc )
			    	goto abort;
			    	
			    rc = store_memory( vax, naddr, ( void * ) &value, 4 );
			    if( rc )
			    	goto abort;
			    	
			}
			
			/* Get length of next fixup list and loop if non-zero */
			
			naddr += 4;
			rc = load_memory( vax, naddr, ( void * ) &fixup_count, 4 );
			if( rc )
				goto abort;
		}
		
	
	}

//	At this point, we can free up the shared image list if we made one.

	for( shr = shrlist; shr; shr = shrnext ) {
		shrnext = shr-> next;
		freemem( shr );
	}
	
//	Set the mode back to what it was when we started, probably USER mode...

	set_mode_stack( vax, saved_mode );

//	Now that we've loaded it all in memory, set up a symbol that defines the entry point.
//	For now, we look for the entry point in user mode, and ignore the LIB$INITIALIZE
//	or debugger entry points.

	addr = 0L;
	if( transfer[ 0 ] > 0L && transfer[ 0 ] < 0x3FFFFFFF )
		addr = transfer[ 0 ];
	else
	if( transfer[ 1 ] > 0L && transfer[ 1 ] < 0x3FFFFFFF )
		addr = transfer[ 1 ];
	else
	if( transfer[ 2 ] > 0L && transfer[ 2 ] < 0x3FFFFFFF )
		addr = transfer[ 2 ];
	
	if( debug )
		printf( "Transfer address is %08lx\n", addr );

//	Well, the user did say to RUN it, so set up a call to the entry point...

	if( !addr )
		printf( "No transfer address!\n" );
	else {
	
		strcpy( sb, "MAIN" );
		sbp = sb;
        rc = set_symbol( vax, &sbp, addr );
        if( rc != VAX_OK )
        	goto abort;
        vax-> console.last_symbol-> flags |= SYM_ENTRY;

		sprintf( sb, "MAIN", addr );
		sbp = sb;
		return console_call( &vax, &sbp );
	}


	return rc;
	
abort:
	fclose( fp );
	printf( "Error reading image header\n" );
	set_mode_stack( vax, saved_mode );
    return rc;

}



void init_ihd_maps( void )
{

	struct IHD * ihd;
	struct ISD * isd;
	struct IHI * ihi;
	struct IAF * iaf;
	
	if( find_map( "IHD" ))
		return;
	
	ihd = 0L;
	isd = 0L;
	ihi = 0L;
	iaf = 0L;
	
	map_init( "IAF" );
	
	map_add( "IAF", "OFFSET_G_FIX",
				0x0C, STROFF( iaf, offset_g_fix ),
				4, 0 );

	map_add( "IAF", "OFFSET_SHL",
				0x18, STROFF( iaf, offset_shl ),
				4, 0 );

	map_add( "IAF", "OFFSET_ADDR",
				0x10, STROFF( iaf, offset_addr ),
				4, 0 );

	map_add( "IAF", "OFFSET_CHGPROT",
				0x14, STROFF( iaf, offset_chgprot ),
				4, 0 );

	map_add( "IAF", "SHRIMGCNT",
				0x1C, STROFF( iaf, shrimgcnt ),
				4, 0 );


	map_init( "IHI" );
	
	map_add( "IHI", "IMGNAML",
				0, STROFF( ihi, imgnaml ),
				1, 0 );
	
	map_add( "IHI", "IMGNAM",
				1, STROFF( ihi, imgnam ),
				39, 1 );
	
	map_add( "IHI", "IMGIDL",
				40, STROFF( ihi, imgidl ),
				1, 0 );
	
	map_add( "IHI", "IMGID",
				41, STROFF( ihi, imgid ),
				15, 1 );
	
	map_add( "IHI", "TIME_0",
				56, STROFF( ihi, time[ 0 ] ),
				4, 0 );
	
	map_add( "IHI", "TIME_1",
				60, STROFF( ihi, time[ 1 ] ),
				4, 0 );

	map_add( "IHI", "LINKIDL",
				64, STROFF( ihi, linkidl ),
				1, 0 );
	
	map_add( "IHI", "LINKID",
				65, STROFF( ihi, linkid ),
				15, 1 );
	
	
	map_init( "IHD" );
	
	map_add( "IHD", "SIZE", 
				0, STROFF( ihd, size ), 
				2, 0 );
				
	map_add( "IHD", "OFFSET_TRANSFER", 
				2, STROFF( ihd, offset_transfer ), 
				2, 0 );
				
	map_add( "IHD", "OFFSET_DST", 
				4, STROFF( ihd, offset_dst ), 
				2, 0 );
				
	map_add( "IHD", "OFFSET_IDENT", 
				6, STROFF( ihd, offset_ident ), 
				2, 0 );
	
	map_add( "IHD", "OFFSET_PATCH", 
				8, STROFF( ihd, offset_patch ), 
				2, 0 );
	
	map_add( "IHD", "MAJOR_ID", 
				12, STROFF( ihd, major_id ), 
				2, 0 );
	
	map_add( "IHD", "MINOR_ID", 
				14, STROFF( ihd, minor_id ), 
				2, 0 );
	
	map_add( "IHD", "HEADER_BLOCKS", 
				16, STROFF( ihd, header_blocks ), 
				1, 0 );

	map_add( "IHD", "IMAGE_TYPE", 
				17, STROFF( ihd,image_type ), 
				1, 0 );
	
	map_add( "IHD", "MASK_0", 
				20, STROFF( ihd, mask[0] ), 
				4, 0 );
	
	map_add( "IHD", "MASK_1", 
				24, STROFF( ihd, mask[1] ), 
				4, 0 );

	map_add( "IHD", "CHANNELS", 
				28, STROFF( ihd, channels ), 
				2, 0 );
	
	map_add( "IHD", "IO_PAGES", 
				30, STROFF( ihd, io_pages ), 
				2, 0 );
	
	map_add( "IHD", "FLAGS", 
				32, STROFF( ihd, flags ), 
				4, 0 );

	map_add( "IHD", "SECTION_ID", 
				36, STROFF( ihd, section_id ), 
				4, 0 );
	
	map_add( "IHD", "VERSION", 
				40, STROFF( ihd, version ), 
				4, 0 );
	
	map_init( "ISD" );
	
	map_add( "ISD", "SIZE", 
				0, STROFF( isd, size ), 
				2, 0 );

	map_add( "ISD", "PAGES", 
				2, STROFF( isd, pages ), 
				2, 0 );

	map_add( "ISD", "VPN", 
				4, STROFF( isd, vpn ), 
				2, 0 );

	map_add( "ISD", "PFC", 
				6, STROFF( isd, pfc ), 
				2, 0 );

	map_add( "ISD", "FLAGS", 
				8, STROFF( isd, flags ), 
				4, 0 );

	map_add( "ISD", "VBN", 
				12, STROFF( isd, vbn ), 
				4, 0 );

	map_add( "ISD", "SECTION_ID", 
				16, STROFF( isd, section_id ), 
				4, 0 );

	map_add( "ISD", "COUNT", 
				20, STROFF( isd, count ), 
				1, 0 );

	map_add( "ISD", "NAME", 
				21, STROFF( isd, name ), 
				15, 1 );


	return;	
}
