//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     console_show.c
//
//  Purpose:    This module implements the SHOW commands.  These commands
//              display the state of the virtual VAX processor or data about
//              the console processor.
//
//  History:    08/06/97    New header format standardization
//
//              02/01/98    Added SHOW ASM command.  Added printbit formatting routine.
//
//              01/06/99    Add additional ASM mode displays.  Make the
//                          breakpoint display resolve symbols.
//
//              01/07/99    Added SHOW CALLS to display call frames
//
//              01/08/99    Handle "console generated" frames better
//
//              01/11/99    Added SHOW ASM attributes.
//
//              02/01/99    Added SHOW UNIMPLEMENTED INSTRUCTIONS as a tool
//                          to help development.
//
//              02/12/99    Added SHOW PAGE command
//
//              05/03/99    Fix up SHOW STACK so that it stops when it
//                          gets to a page that's protected.  We now
//                          mark the high-address page of each stack as
//                          an inaccessible page.
//
//              05/11/99    Added SHOW STRINGS command that dumps the
//                          assembler's string pool.
//
//				05/25/99	Modified format of SHOW SYS SYM and SHOW SYM
//							to include more symbol format modifiers.
//
//				05/25/99	Added SHOW INTERRUPT
//
//				06/03/99	Added DBG_REGISTERS to the SHOW DEBUG display.
//
//				06/14/99	Added fault breakpoints to SHOW BREAKPOINTS
//
//				07/13/99	Added SHOW MMAP command to dump struct maps.


#include "vax.pch"
#include "console_proto.h"
#include "asmproto.h"
#include "vaxinstr.h"
#include "pte.h"
#include "memmap.h"


extern char * pr_names[];

static char * printbit( long flag, long mask, long normal, 
                                    char * name, char * desc );
static void show_instructions( long match );
long tracevm( struct VAX * vax, unsigned long Addr, short mode );
static long showscb( struct VAX * vax, long all );
void dump_tb( struct VAX * vax );
struct SYMBOL * find_label( struct VAX * vax, long dest, long flags );

extern long tb_hit, tb_try, tb_flush, tb_pflush;


//  The SHOW command.

long console_show( struct VAX ** Vax, char ** P )
{
    long rc, opmatch;
    struct VAX * vax;
    char * p, *pn, *ps;
    long count, n, i, verb, value, mode;
    long all, unimplemented, unresolved, system;
    struct BREAKSTR * breakpoint;
    struct SYMBOL * sym, *sp;
    char buff[ 256 ], m[ 32 ], *bp;
    char ch, match;
    unsigned long addr;
    float ratio;
    long desc_len;
    unsigned long desc_addr, desc;

    long stack_id;
    unsigned long stringpool_base, stringpool_size, stringpool, stringaddr;
    long saved_mode, saved_is;

    static char * mode_names[ 5 ] = {"KERNEL", "EXEC", "SUPER", "USER", 
                                       "INTERRUPT" };
/*
 *  Get local addressability to virtual VAX and command buffer
 */
 
    vax = *Vax;
    p = *P;

    
/*
 *  If there is no active virtual VAX, we are done!
 */
 
    if( vax == 0L ) {
        rc = VAX_NOVAX;
        return rc;
    }

/*
 *  Many of the SHOW commands can display the TODR register explicilty
 *  or implicitly, so let's update it now.  We only do this if the register
 *  is already non-zero in value (VAX 11/750 standard).
 */
        
    if( vax-> TODR != 0L ) {
            /* Difference is 60th's of a second, make seconds them 1000's */
        vax-> TODR = (( TickCount() - vax-> timebase ) * 100 ) /60;
    }
                
	vax-> ICR = vax-> clock;

/*
 *  See if it's a SHOW command for a privileged register name.
 */

    //  Copy the name to a local buffer.
    
    n = 0;
    flush_blanks( &p );
    while( !isend( vax, *p ) && !is_blank( *p )) {
    
        ch = *(p++);
        if( ch >= 'a' && ch <= 'z' )
            ch = ( char ) ( ch - 32 );
        buff[ n++ ] = ch;
        
    }
    buff[ n ] = 0;

    //  Search the list of privileged registers to see if it's a match
    
    match = 0;
    for( n = 0; n < MAXPRIVREG; n++ ) {
    
        if( strcmp( buff, pr_names[ n ] ) == 0 ) {
            match = 1;
            break;
        }
    }

	//  Note that we make KSP, ESP, SSP, USP, and ISP special cases, which are handled
	//  later as special versions of the SHOW STACK function.

	if( n <= 4 )
		match = 0;

    //  If a match was found, it's a privileged register and we're done.
        
    if( match ) {
        *P = p;
        printf( "   PREG %6s = %08lX (hex)  %12ld (dec)\n", pr_names[ n ], 
			vax-> preg[ n ], vax-> preg[ n ] );
        return VAX_OK;
    }



#if 0   /* I think we want to force them to use the SYMBOL sub-verb */

/*
 *  See if it's a SHOW <symbol> command for a known symbol.  We must set the NOFORWARD
 *  flag here to make sure that the get_symbol function won't _create_ a symbol given
 *  the name we specify.
 */
 
    vax-> assembler.displacement = K_NOFORWARD;
    
    bp = buff;
    rc = get_symbol( vax, &bp, ( long * ) &value );
    if( rc == VAX_OK ) {
    
        if( dump_forward_references( vax, vax-> console.last_symbol ) == 0 )
            printf( "   SYM %s = %08lX (hex)   %12ld (dec)\n", vax-> console.last_symbol-> name, value, value );
        *P = p;
        return rc;
    }
#endif
    

/*
 *  It wasn't a symbol, so reset to the start again and read off the first verb token.
 */
 
    p = *P;
    
    rc = read_verb( vax, &p, &verb );
    if( rc != VAX_OK )
        return rc;
    *P = p;
    
/*
 *  See if it's a request to examine a register by name 
 */
 
    rc = exam_reg( vax, verb );
    if( rc == VAX_OK )
        goto exit;
    rc = VAX_OK;
    

/*
 *  See if it's the modifier ALL which can be used on some commands.
 */

    if( verb == CHAR4('A','L','L',' ') || verb == CHAR4('/','A','L','L')) {
        all = 1;
        read_verb( vax, &p, &verb );
        *P = p;
    }
    else
        all = 0;

/*
 *  See if it's the UNRESOLVED modifier, used by SHOW SYMBOLS.
 */

    if( verb == CHAR4('U','N','R','E') ) {
        all = 1;
        unresolved = 1;
        read_verb( vax, &p, &verb );
        *P = p;
    }
    else
        unresolved = 0;

/*
 *  See if it's a SYSTEM prefix.
 */

    if( verb == CHAR4('S','Y','S','T') || verb == CHAR4('S','Y','S',' ') ) {
        system = 1;
        read_verb( vax, &p, &verb );
        *P = p;
    }
    else
        system = 0;

/*
 *  See if it's the UNIMPLEMENTED prefix to INSTRUCTIONS
 */

    if( verb == CHAR4('U','N','I','M') ) {
        unimplemented = 1;
        read_verb( vax, &p, &verb );
        *P = p;
    }
    else
        unimplemented = 0;

/*
 *  Dispatch based on the parameter verb.
 */
 
    switch( verb ) {

    case CHAR4('S','T','R','I'):
	case CHAR4('S','T','R',' '):
        strcpy( buff, "CONSOLE$STRINGPOOL" );
        bp = buff;
        rc = get_symbol( vax, &bp, ( long * ) &stringpool );
        if( rc )
            return rc;

        strcpy( buff, "CONSOLE$STRINGPOOL_BASE" );
        bp = buff;
        rc = get_symbol( vax, &bp, ( long * ) &stringpool_base );
        if( rc )
            return rc;

        strcpy( buff, "CONSOLE$STRINGPOOL_SIZE" );
        bp = buff;
        rc = get_symbol( vax, &bp, ( long * ) &stringpool_size );
        if( rc )
            return rc;

        printf( "STRING POOL\n" );
		printf( "    CONSOLE$STRINGPOOL_BASE       %08lX\n", stringpool_base );
		printf( "    CONSOLE$STRINGPOOL_SIZE       %08lX\n", stringpool_size );
		printf( "    CONSOLE$STRINGPOOL (Current)  %08lX\n", stringpool );

        rc = load_memory( vax, stringpool_base, ( void * ) &stringaddr, 4 );

        if( stringaddr )
            printf( "\n" );

        while( stringaddr ) {

            rc = load_memory( vax, stringaddr, ( void * ) &desc_len, 4 );
            if( rc )
                return rc;
            desc_len = desc_len & 0x0000FFFF;

            rc = load_memory( vax, stringaddr+4, ( void * ) &desc_addr, 4 );
            if( rc )
                return rc;

            desc = stringaddr;
            rc = load_memory( vax, stringaddr+8, ( void * ) &stringaddr, 4 );
            if( rc )
                return rc;

            if( desc_len <= 250 )
                buff[ desc_len ] = 0;

            for( n = 0; n < desc_len; n++ ) {
                if( n > 250 ) {
                    buff[ 250 ] = 0;
                    strcat( buff, "..." );
                    break;
                }
                rc = load_memory( vax, desc_addr+n, 
                          ( void *) &( buff[ n ]), 1 );
                if( rc )
                    return rc;
            }

            printf( "    Descriptor %08lX [%3ld bytes]  \"%s\"\n",
                    desc, desc_len, buff );
        }

		printf( "\n" );

        break;


    case CHAR4('E','R','R',' '):
    case CHAR4('E','R','R','O'):
    
        flush_blanks( &p );
        if( isend( vax, *p )) {
            strcpy( buff, "$STATUS" );
	        bp = buff;
	        get_symbol( vax, &bp, &value );
	        printf( "Last error code was %08lX, \"%s\"\n", value, vaxmsg( value ));
	    }
	    else {
	        rc = asm_expr( vax, &p, &value );
	        if( rc )
	            return rc;
            printf( "Error code %08lX, \"%s\"\n", value, vaxmsg( value ));
        }
        break;
        
    case CHAR4('M','O','D','E'):

        printf( "Current MODE is %s\n", mode_names[ vax-> pslw.cur_mod ] );
        break;

    case CHAR4('P','T','E',' '):
    case CHAR4('P','A','G','E'):

        rc = asm_expr( vax, &p, &value );
        if( rc )
            return rc;

        flush_blanks( &p );
        rc = read_verb( vax, &p, &verb );
        if( rc == VAX_OK )
           mode = ( verb == CHAR4('W','R','I','T') || verb == CHAR4('W',' ',' ',' ') );
        else
           mode = 0;

        rc = tracevm( vax, ( unsigned long ) value, ( short ) mode );
        break;
        
    case CHAR4('T','B',' ',' '):

        printf( "Translation buffer caching is %s\n",
          vax-> TBDR ? "disabled" : "enabled" );

        if( tb_try == 0L ) 
            ratio = 0.0;
        else {
            ratio = ( float ) tb_hit;
            ratio = ( float ) ( ratio / ( float ) tb_try * 100.0 );
        }
        printf( "    Tries=%d     Hits=%d    Ratio=%d%%\n",
            tb_try, tb_hit, ( long ) ratio );
        printf( "    Flushes=%ld   PFlushes=%ld\n", tb_flush, tb_pflush );
        
        dump_tb( vax );

        break;

    case CHAR4('C','A','L','L'):

        flush_blanks( &p );
        if( !isend( vax, *p )) {
            rc = asm_expr( vax, &p, &value );
            if( rc )
                return rc;
        }
        else
            value = 255;

        rc = show_calls( vax, value );
        break;

    case CHAR4('Q','U','A','N'):    

        printf( "QUANTUM\n    INTERRUPTS  Initial=%ld  Current=%ld\n",
                vax-> quantum.initial, vax-> quantum.current );
        
        printf( "    USER INTF   Initial=%ld  Current=%ld\n",
                vax-> uiquantum.initial, vax-> uiquantum.current );

        break;

	case CHAR4('M','A','P',' '):
	case CHAR4('M','M','A','P'):
	case CHAR4('M','A','P','S'):
	
		map_dump();
		break;
		
    case CHAR4('D','E','B','U'):
    case CHAR4('D','B','G',' '):
    
        printf( "DEBUG SETTINGS:\n" );
        
        printf( "    %s\n", printbit( vax-> debug,
                                    DBG_DEBUG,
                                    0, 
                                    "DEBUG",
                                    "Invoke native debugger?" ));

        printf( "    %s\n", printbit( vax-> debug,
                                    DBG_VM,
                                    0, 
                                    "VM",
                                    "Debug virtual memory translations?" ));

        printf( "    %s\n", printbit( vax-> debug,
                                    DBG_TB,
                                    0, 
                                    "TB",
                                    "Debug translation buffer caching?" ));

        printf( "    %s\n", printbit( vax-> debug,
                                    DBG_SYMBOLS,
                                    0, 
                                    "SYMBOLS",
                                    "Debug symbol table handling?" ));

        printf( "    %s\n", printbit( vax-> debug,
                                    DBG_EXCEPTIONS,
                                    0, 
                                    "EXCEPTIONS",
                                    "Debug exception handling?" ));

		printf( "    %s\n", printbit( vax-> debug,
									DBG_INTERRUPTS,
									0,
									"INTERRUPTS",
									"Debug interrupt handling?" ));

		printf( "    %s\n", printbit( vax-> debug,
                                    DBG_REGISTERS,
                                    0, 
                                    "REGISTERS",
                                    "Display changed registers on STEP?" ));

		printf( "    %s\n", printbit( vax-> debug,
                                    DBG_IMAGES,
                                    0, 
                                    "IMAGES",
                                    "Display image info on RUN command?" ));


        break;

        
    case CHAR4('A','S','S','E'):
    case CHAR4('A','S','M',' '):
    
        printf( "ASSEMBLER SETTINGS:\n" );
        
        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_ADDRPROMPT, 
                                    0, 
                                    "ADDRESS_PROMPT",
                                    "Include address in ASM prompt?" ));
        
        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_WARNFORWARD,
                                    0,
                                    "FORWARD_WARNINGS",
                                    "Warn if unresolved references after ASM END" ));
        
        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_BRANCHDEST,
                                    0,
                                    "BRANCH_DESTINATION",
                                    "Display branch destination addresses" ));

        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_SYMBOLS,
                                    0,
                                    "SYMBOLS",
                                    "Display addresses symbolically" ));

        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_RESTMP,
                                    0,
                                    "RESOLVE_TEMP",
                                    "Require temporary sym resolution in scope"));

        printf( "    %s\n", printbit( vax-> assembler.flags,
                                    ASM_SCOPENAMES,
                                    0,
                                    "SCOPENAMES",
                                    "Use scope names in temp symbols" ));

        printf( "\n" );
        
        break;
        
            
    case CHAR4('I','N','S','T'):
    

        count = 0;
        flush_blanks( &p );
        if( !isend( vax, *p )) {

           /* See if this is SHOW INSTRUCTION PROFILE */

            ps = p;
            read_verb( vax, &ps, &verb );
            if( verb == CHAR4('P','R','O','F') ) {
                *P = ps;
                show_instructions( -2 );
                return VAX_OK;
            }

           /* No, see if its a particular opcode to display */

            all = 1;
            rc = asm_hex( vax, &p, ( void * ) &opmatch );
            if( rc != VAX_OK )
                return rc;
            
            *P = p;
        }
        else
            opmatch = -1;
            
        if( !all || opmatch == -1 )
            printf( "INSTRUCTION OPCODES:\n\n" );
        
        /* The default is to just show the opcode values */
        
        if( !all ) {
            n = 0;
            for( i = 0;  i < 500; i++ ) {
                pn = instruction[ i ].name;
                
                if( *pn == 0L )
                    break;

                /* IF the 'unimplemented' flag is set, then the routine */
                /* must not be implemented, else loop again.            */

                if( unimplemented && instruction[ i ].routine != 0L )
                    continue;

                /* Likewise if the flag is not set then the routine MUST */
                /* be implemented or we loop again.                      */

                if( !unimplemented && instruction[ i ].routine == 0L )
                    continue;

				/* If it's not an instruction, skip it... */

				if( strncmp( instruction[ i ].name, "RSVD_", 5 ) == 0 )
					continue;
				if( strncmp( instruction[ i ].name, "EXT_", 4 ) == 0 )
					continue;

                count++;
                if( instruction[ i ].extended )
                    printf( "   %02X %02X %-8s  ",
                        instruction[ i ].extended,
                        instruction[ i ].opcode,
                        instruction[ i ].name );
                else
                    printf( "      %02X %-8s  ",
                        instruction[ i ].opcode,
                        instruction[ i ].name );
                n++;
                if( n == 4 ) {
                    printf( "\n" );
                    n = 0;
                }
            }
            if( n )
                printf( "\n" );
			printf( "\n%d Instructions.\n", count );

        }
        else
            show_instructions( opmatch );
            
        break;
                
    case CHAR4('B','R','E','A'):
    case CHAR4('B','R',' ',' '):
    
        if( !vax-> console.breakpoint_list ) {
            printf( "No breakpoints defined\n" );
            break;
        }
        
        printf( "    Breakpoints:\n" );
        for( breakpoint = vax-> console.breakpoint_list; breakpoint; breakpoint = breakpoint-> next ) {

            buff[ 0 ] = 0;
            if( breakpoint-> kind == BREAK_FAULT ) {
            	sprintf( buff, "%s", vaxexcept(( short ) breakpoint-> pc ));
            }
            else {
	            for( sp = vax-> console.symbols; sp; sp = sp-> next ) {
	
					if((( unsigned long ) sp-> value == ( unsigned long ) breakpoint-> pc - 2  ) &&
						( sp-> flags & SYM_ENTRY ))
						break;
	
	                if((( unsigned long ) sp-> value == ( unsigned long ) breakpoint-> pc ) &&
	                   ( sp-> flags & SYM_LABEL ))
	                    break;
				}
	
				if( !sp )
	            for( sp = vax-> console.system_symbols; sp; sp = sp-> next ) {
	
					if((( unsigned long ) sp-> value == ( unsigned long ) breakpoint-> pc - 2  ) &&
						( sp-> flags & SYM_ENTRY ))
						break;
	
	                if((( unsigned long ) sp-> value == ( unsigned long ) breakpoint-> pc ) &&
	                   ( sp-> flags & SYM_LABEL ))
	                    break;
				}
				
	            if( sp )
	                strcpy( buff, sp-> name );
            }
			if( breakpoint-> after )
	            printf( "  %c %08lX (%3d) %s\n", 
					(breakpoint-> kind == BREAK_FAULT) ? 'F' : ' ',
					breakpoint-> pc,
						breakpoint-> after, buff );
			else
	            printf( "  %c %08lX       %s\n", (breakpoint-> kind == BREAK_FAULT) ? 'F' : ' ',
						breakpoint-> pc, buff );

        }
        break;
        
    case CHAR4('R','E','G','I'):
    case CHAR4('R','E','G','S'):
    case CHAR4('R','E','G',' '):

        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }

        rc = dump_registers( vax );
        break;
    
    case CHAR4('P','S','L',' '):

        rc = dump_psl( vax );
        break;
        
    case CHAR4('C','P','U',' '):
    
        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }
        
        rc = dump_registers( vax );
        if( rc )
            break;
        
        printf( "\n    Stack pointers:\n\n" );
        printf( "    USP: %08lX   SSP: %08lX   ESP: %08lX   KSP: %08lX\n    ISP: %08lX\n\n",
                    vax-> USP, vax-> SSP, vax-> ESP, vax-> KSP,
                    vax-> ISP );
        
        //  Print the privileged registers as a block.  Each time the output buffer
        //  line gets long enough, dump it out.
        
        strcpy( buff, "    " );
        
        printf( "    Privileged registers:\n\n" );
        for( n = 5; n < MAXPRIVREG; n++ ) {
            pn = pr_names[ n ];
            if( *pn == '_' )
                continue;
            
            sprintf( m, "    %-6s:  %08lX", pn, vax-> preg[ n ] );
            strcat( buff, m );
            if( strlen( buff ) > 60 ) {
                strcat( buff, "\n" );
                printf( buff );
                strcpy( buff, "    " );
            }   
        }
        
        //  If there's stuff left in the output buffer, flush it now.  We check the
        //  [1] character because it would be null if the buffer was only initialized
        //  to a tab character.
        if( buff[ 1 ] ) {
            strcat( buff, "\n" );       
            printf( buff );
        }
    
        
    case CHAR4('D','E','P','O'):
    case CHAR4('B','A','S','E'):

        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }
    
        printf( "\n        Next storage address is %08lX\n", vax-> console.deposit );
        break;
    
    case CHAR4('V','M',' ',' '):
    case CHAR4('M','A','P','E'):
    case CHAR4('M','E','M',' '):
    case CHAR4('M','E','M','O'):

        flush_blanks( &p );
        if( !isend( vax, *p )) {

           /* See if this is SHOW MEMORY STAT debugging command */
           
            ps = p;
            read_verb( vax, &ps, &verb );
            if( verb == CHAR4('S','T','A','T') ) {
                *P = ps;
                printmem();
                return VAX_OK;
            }
        }

        show_regions( vax );
            
        break;
    
	case CHAR4('K','S','P',' '):
	case CHAR4('U','S','P',' '):
	case CHAR4('S','S','P',' '):
	case CHAR4('E','S','P',' '):
	case CHAR4('I','S','P',' '):
    case CHAR4('S','T','A','C'):
    
		saved_mode = 0L;


		if( verb != CHAR4('S','T','A','C') ) {
			saved_mode = vax-> pslw.cur_mod;
			saved_is   = vax-> pslw.is;

			/* Because a mode switch can happen, dump TB protection cache */
            invalidate_tb_prot(); 
			switch( verb ) {

			case CHAR4('K','S','P',' '):  set_mode_stack( vax, 0 ); break;
			case CHAR4('E','S','P',' '):  set_mode_stack( vax, 1 ); break;
			case CHAR4('S','S','P',' '):  set_mode_stack( vax, 2 ); break;
			case CHAR4('U','S','P',' '):  set_mode_stack( vax, 3 ); break;
			case CHAR4('I','S','P',' '):  set_mode_stack( vax, 4 ); break;
			}
		}

        flush_blanks( &p );
        if( isend( vax, *p ))
            count = all ? 
                  ( vax-> preg[ vax-> pslw.cur_mod ] - vax-> SP ) >> 2 : 1;
        else {
            rc = asm_hex( vax, &p, ( unsigned long * ) &count );
            if( rc )
                goto stack_exit;
        }
        
        if( vax-> pslw.is )
            stack_id = 4;
        else
            stack_id = vax-> pslw.cur_mod;

        printf( "    %s MODE SP = %08lX:\n", 
                   mode_names[ stack_id ], vax-> SP );
        
        count = count * 4;
        
        for( n = 0; n < count; n += 4 ) {
        
            addr = vax-> SP + n;
            
            /* If virtual memory is on, then the stack top is end of P1 */
            /* if we are in USER mode.  If virtual memory is off, then  */
            /* it's the end of physical memory. */
            
            if(( vax-> pslw.cur_mod == 3 && vax-> MAPEN && addr >= 0x7FFFFFFF ) || 
               ( vax-> MAPEN == 0 && ( addr >= ( unsigned long ) vax-> memsize ))) {
                printf( "      <end of memory>\n" );
                goto stack_exit;
            }

            if( probe( vax, addr, 4, VM_READ | VM_NOSIGNAL )) {
                printf( "      <end of stack>\n" );
                goto stack_exit;
            }
            
            rc = load_memory( vax, addr, ( unsigned char * ) &value, 4 );
			if( rc )
                goto stack_exit;
                
            if( vax-> console.radix == 16 ) 
                printf( "      SP+%04lX [%08lX]:  %08lX\n", n, addr, value );
            else
                printf( "      SP+%04ld [%08lX]:  %ld\n", n, addr, value );

        }
        
stack_exit:	if( verb != CHAR4('S','T','A','C') ) {
				set_mode_stack( vax, saved_mode );
			}

        break;
        
    case CHAR4('E','X','C','E'):
    case CHAR4('E','X','C',' '):
    case CHAR4('I','N','T','E'):
    case CHAR4('I','N','T',' '):
    
        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }
        
        if( vax-> fault.code == 0 ) {
            printf( "    No exceptions or interrupts have occurred yet\n" );
        }
		else {        
	        printf( "    Last exception/interrupt code was %08lX, %s\n        %ld argument%s,", 
	                vax-> fault.code, vaxexcept(( short ) vax-> fault.code ),
	                    vax-> fault.signal_args[ 0 ],
	                    vax-> fault.signal_args[ 0 ] == 1 ? "" : "s" );
	                    
	        printf( "    PC=%08lX   PSL=%08lX\n",
	                vax-> fault.pc, vax-> fault.psl );
	        
	        
	        for( n = 1; ( unsigned long ) n <= vax-> fault.signal_args[ 0 ]; n++ ) {
	            printf( "        Argument %ld is %08lX\n", n, vax-> fault.signal_args[ n ] );
	        }
		}
		        
        if( vax-> interrupt_pending ) {
        	printf( "    There is a pending interrupt code %s at ipl %02X\n",
        				vaxexcept(( short ) vax-> interrupt_pending) , 
        				vax-> interrupt_ipl );
        }
        
        if( vax-> iqueue ) {
        
        	struct INTERRUPT * ip;
        	
        	printf( "Pending interrupts:\n" );
        	printf( "    IPL  Quantum  Age Code\n" );

  		    for( ip = vax-> iqueue; ip; ip = ip-> next ) {
        		printf( "    %02lX   %4ld   %4ld  %02lX %s\n",
        				ip-> ipl, ip-> quantum, ip-> age, ip-> code, 
        				vaxexcept((short) ip-> code ));
        	}
        }
        
        break;
        
    case CHAR4('R','A','D',' '):
    case CHAR4('R','A','D','I'):
        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }
        printf( "    Default radix is %d\n", vax-> console.radix );
        break;
    
    case CHAR4('D','I','S','A'):
        if( all ) {
            rc = VAX_SYNTAX;
            break;
        }
        printf( "    Default disassembly flag is %d\n", vax-> console.disasm );
        break;

    case CHAR4('S','C','B',' '):
        return showscb( vax, all );
    
    case CHAR4('S','Y','M',' '):
    case CHAR4('S','Y','M','B'):

        if( system ) {
            rc = dump_system_symbols( vax );
        }
        else
        if( all ) {
        
            /* flag means dump all or dump unresolved... */
            
            rc = dump_symbols( vax, unresolved );
            
            if( rc ==  VAX_ASMUNDSYMS )
                rc = VAX_OK;
                
        }
        else {
            
			char attr[ 60 ];

            rc = get_symbol( vax, &p, ( long * ) &value );
            if( rc )
                break;
            
            sym = vax-> console.last_symbol;

			attr[ 0 ] = 0;

			if( sym != 0L ) {
				if( sym-> flags & SYM_PERMANENT )
					strcat( attr, "perm " );
				if( sym-> flags & SYM_ENTRY )
					strcat( attr, "entry " );
				if( sym-> flags & SYM_LABEL )
					strcat( attr, "label " );
				if( sym-> flags & SYM_LOCAL )
					strcat( attr, "local " );
			}
            
            /* Dump the forward references.  If there are none print the value normally */
            
            if( dump_forward_references( vax, sym ) == 0 )
                printf( "    %s = %08lX (hex)   %12ld (dec) %s\n", 
				    sym-> name, value, value, attr );
            
            ;
        }
        
        break;
        
    default:
        rc = VAX_UNKPARM;
        break;
        
    }
    
    
/*
 *  Store parameters back to caller.
 */

exit:

    *Vax = vax;
    *P = p;
    return rc;
}


//  Utility routine.  Given a flag and mask bit, format for output.

char * printbit( long flag, long mask, long no_value, char * name, char * desc )
{
    static char * msg = 0L;
    
    char n[ 32 ];
    long value;
    
    if( msg == 0L )
        msg = getmem( 256 );
    
    if( msg == 0L )
        return "<insufficient memory to format data>";

    value = flag & mask;
    if( value )
        value = 1;
    
    if( value != no_value )
        strcpy( n, name );
    else {
        strcpy( n, "NO" );
        strcat( n, name );
    }
    
    if( desc != 0L )
        sprintf( msg, "%-20s    %s", n, desc );
    else
        sprintf( msg, "%s", n );
        
    return msg;
}



/*
 *  show_instructions()
 *
 *  This does the SHOW ALL INSTRUCTION command, which dumps each opcode with
 *  lots of formatted information about how the instruction works.
 */

void show_instructions( long match )
{

    short i, size, j;
    char * pn;
    char m[ 200 ];
    char b[ 32 ];
    long access;
    long found;
    long count;

/*  Loop over the instruction data store */

    found = 0;
    count = 0;
    for( i = 0; i < 500; i++ ) {
    
        /* If the name is empty then we've hit the end of the list */
        
        pn = instruction[ i ].name;
        if( *pn == 0 )
            break;
    
        /* If it's not implemented, skip it */
        
        if( instruction[ i ].routine == 0L )
            continue;
        
        /* If we want a particular one, test for it */
        
        if( match >= 0 && instruction[ i ].opcode != match )
            continue;
        else
            found = 1;
            
        
        /*  Start the output buffer with the usage count data */
        /*  if we are asking for profile data (match = -2)    */
        /*  If we show profile data and the count is zero     */
        /*  then don't display it.                            */

        if( match == -2 ) {
            if( instruction[ i ].use_count )
                sprintf( m, "(%5ld) ", instruction[ i ].use_count );
            else
                continue;
        }
        else
            m[ 0 ] = 0;


        /*  Format the opcode values and the name at the start of the buffer */
        
        if( instruction[ i ].extended )
            sprintf( b, "%02X %02X %-8s",   instruction[ i ].extended,
                                            instruction[ i ].opcode,
                                            pn );
        else
            sprintf( b, "   %02X %-8s",     instruction[ i ].opcode,
                                            pn );
        
        strcat( m, b );
        
        /*  Loop over the operand list */
        
        for( j = 0; j < instruction[ i ].operand_count; j++ ) {
        
            access = instruction[ i ].access;
            access = GETMODE( access, j );
            
            size = instruction[ i ].scale[ j ];
                
            if( j > 0 )
                strcat( m, ", " );
            
            /* Build the operand mode and type */
            
            switch( access ) {
            
            case OP_RD: pn = "src";     break;
            case OP_WR: pn = "dst";     break;
            case OP_MD: pn = "mod";     break;
            case OP_AD: pn = "addr";    break;
            case OP_VA: pn = "var";     break;
            case OP_BR: pn = "br";      break;
            default:    pn = "x";       break;
            }
            strcat( m, pn );
            
            switch( size ) {
            case 1: pn = ".b";  break;
            case 2: pn = ".w";  break;
            case 4: pn = ".l";  break;
            case 8: pn = ".q";  break;
            default:    pn = ".x";  break;
            }
            
            strcat( m, pn );
            
        }
        printf( "    %s\n", m );
        count++;
    }
    
    if( match == -1 && count > 1 )
        printf( "%ld instructions\n", count );
        
    if( match >= 0 && !found ) 
        printf( "No implemented instruction for opcode %02lX\n", match );

    return;
}




/*
 *  SHOW_REGIONS
 *
 *  This implements the code that dumps out the VM table data.  It's called from VMINIT
 *  to show what it did, and from SHOW VM when virtual memory is enabled.
 */

long show_regions( struct VAX * vax )
{

    long n, base, len;

    if( vax == 0L )
        return VAX_NOVAX;

//  Dump what we know about physical memory.

    len = vax-> memsize / 512;

    printf( "\n    Physical Memory\n        %08lx (%ld decimal) pages\n        Addresses  00000000 - %08lX\n\n",
            len, len, vax-> memsize - 1 );
            
//  If the VMINIT command has never been issued, then we don't have console-specific
//  knowledge of the memory layout, so we can't report on it.  Note that VAX software
//  may well have set up VM status, but _we_ don't know about it.

    printf( "    Virtual Memory (currently %s)\n",
        vax-> MAPEN ? "ENABLED" : "DISABLED" );

    if( !vax-> vm_initialized ) {
        printf( "        Virtual memory is not initialized.\n" );
        
        return VAX_OK;
    }

    
//  Loop over each of the regions and tell what we know about it.

    base = 0L;
	len = 0L;

    for( n = 0; n < 3; n++ ) {
    
        if( n == 0 ) {
            base = vax-> P0BR;
            len  = vax-> P0LR;
        }
        else
        if( n == 1 ) {
            base = vax-> P1BR;
            len  = vax-> P1LR;
        }
        else
        if( n == 2 ) {
            base = vax-> SBR;
            len  = vax-> SLR;
        }

        printf( "        %s Region\n            Region size = %08lX (%5ld decimal) pages\n            PFN database = %ld page%s",
            vax-> region[ n ].name, 
            vax-> region[ n ].size, 
            vax-> region[ n ].size, 
            vax-> region[ n ].pte_count,
            ( vax-> region[ n ].pte_count == 1 ) ? " " : "s" );

        printf( "  %sBR = %08lX    %sLR = %08lX\n",
            vax-> region[ n ].name, base, vax-> region[ n ].name, len );
            
        printf( "            Physical addresses = %08lX-%08lX\n            Virtual addresses  = %08lX-%08lX\n\n",
            vax-> region[ n ].p_start, vax-> region[ n ].p_end -1,
            vax-> region[ n ].v_start, vax-> region[ n ].v_end -1 );
        
    }

    
    return VAX_OK;
}

//
//  Utility routine, formats the stack and shows the call frame(s)
//

long show_calls( struct VAX * vax, long count )
{

        unsigned long fp;
        long v;

        union MASKREG {
            long longword;
            struct {
                int   spa: 2 ;
                int   calltype: 1 ;
                int   mbz: 1 ;
                int   mask: 12 ;
                int   psw: 16 ;
            } bits;
        } mask_union;
        unsigned long mask, addr, ap, n, saved_ap, rc, argc;


        fp = vax-> FP;
        ap = vax-> AP;

        if( fp == 0L || ap == 0L ) {
            rc = VAX_NOFRAMES;
            return rc;
        }

        while( count > 0 ) {

            count--;
            printf( "    FRAME: %08lX", fp );
            addr = fp;

            /* Get the condition handler address */

            rc = load_memory( vax, addr, ( void * ) &v, 4 );
            if( rc )
                return rc;

            if( v != 0L )
                printf( ", Handler: %08lX", v );

            /* Get the mask longword that defines the frame format */

            addr += 4;
            rc = load_memory( vax, addr, ( void * ) &mask_union.longword, 4 );
            if( rc )
                return rc;

            printf( ", SPA: %1X, %s, MASK: %04X, PSW: %04X\n",
                      mask_union.bits.spa,
                      mask_union.bits.calltype == 0 ? "CALLG" : "CALLS",
                      mask_union.bits.mask,
                      mask_union.bits.psw );


            /* Get the PC, FP, and PC registers */

            addr += 4;
            rc = load_memory( vax, addr, ( void * ) &v, 4 );
            if( rc )
                return rc;
            printf( "        Saved AP : %08lX\n", v );
            saved_ap = v;

            addr += 4;
            rc = load_memory( vax, addr, ( void * ) &v, 4 );
            if( rc )
                return rc;

            if( v == 0xFFFFDEAF )
                printf( "        Saved FP : <console>\n" );
            else
                printf( "        Saved FP : %08lX\n", v );
            fp = v;

            addr += 4;
            rc = load_memory( vax, addr, ( void * ) &v, 4 );
            if( rc )
                return rc;

            if( v == 0xFFFFDEAF )
                printf( "        Saved PC : <console>\n" );
            else
                printf( "        Saved PC : %08lX\n", v );

            /* Get the saved registers, if any */

            addr += 4;
            mask = mask_union.bits.mask;
            for( n = 0; n <= 11; n++ ) {

                if( mask & ( 1 << n )) {
                    rc = load_memory( vax, addr, ( void * ) &v, 4 );
                    if( rc )
                        return rc;
                    addr += 4;
                    printf( "        Saved R%-2ld: %08lX\n", n, v );
                }
            }

            /* Get the argument count */

            rc = load_memory( vax, ap, ( void * ) &argc, 4 );
            if( rc )
                return rc;

            if( argc > 0 )
                printf( "        There %s %ld argument%s:\n",
                     argc == 1 ? "is" : "are",
                     argc,
                     argc == 1 ? "" : "s" );

            if( argc > 0 ) {
                addr = ap;
                for( n = 1; n <= argc; n++ ) {
                    addr += 4;
                    rc = load_memory( vax, addr, ( void * ) &v, 4 );
                    if( rc )
                        return rc;
                    printf( "            [%2ld]: %08lX\n", n, v );
                }
            }
            ap = saved_ap;

        /* The value of FP was captured above and will be the new */
        /* frame pointer now.  However, if the FP is zero then we */
        /* are done. Also if the FP is the magic value 'fdef' we  */
        /* are looking at a frame created by the console and are  */
        /* unable to continue.                                    */

            if( fp == 0L || fp == 0xFFFFDEAF )
                break;

            printf( "\n" );

        }

        return VAX_OK;
}

//
//   Show the SCB entries that are non-zero
//

static long showscb( struct VAX * vax, long all )
{

    unsigned long addr, vector, stack;
    long n, rc, count, saved_mapen;

    struct SYMBOL * sp;

    static char * vector_name[] = {
    /* 00  */  "UNUSED",
    /* 04  */  "CHECK",
    /* 08  */  "KSNV",
    /* 0C  */  "POWER",
    /* 10  */  "PRIV",
    /* 14  */  "CUSTOMER",
    /* 18  */  "RESOP",
    /* 1C  */  "RESADDR",
    /* 20  */  "ACCVIO",
    /* 24  */  "TNV",
    /* 28  */  "TP",
    /* 2C  */  "BPT",
    /* 20  */  "COMPAT",
    /* 34  */  "ARITH",
    /* 38  */  "RESERVED38",
    /* 3C  */  "RESERVED3C",
    /* 40  */  "CHMK",
    /* 44  */  "CHME",
    /* 48  */  "CHMS",
    /* 4C  */  "CHMU",
    /* 50  */  "SBI",
    /* 54  */  "CMRD",
    /* 58  */  "SBIALERT",
    /* 5C  */  "SBIFAULT",
    /* 60  */  "MWT",
    /* 64  */  "RESERVED64",
    /* 68  */  "RESERVED68",
    /* 6C  */  "RESERVED6C",
    /* 70  */  "RESERVED70",
    /* 74  */  "RESERVED74",
    /* 78  */  "RESERVED78",
    /* 7C  */  "RESERVED7C",
    /* 80  */  "RESERVED80",
    /* 84  */  "SOFTWARE1",
    /* 88  */  "SOFTWARE2",
    /* 8C  */  "SOFTWARE3",
    /* 90  */  "SOFTWARE4",
    /* 94  */  "SOFTWARE5",
    /* 98  */  "SOFTWARE6",
    /* 9C  */  "SOFTWARE7",
    /* A0  */  "SOFTWARE8",
    /* A4  */  "SOFTWARE9",
    /* A8  */  "SOFTWARE10",
    /* AC  */  "SOFTWARE11",
    /* B0  */  "SOFTWARE12",
    /* B4  */  "SOFTWARE13",
    /* B8  */  "SOFTWARE14",
    /* BC  */  "SOFTWARE15",
    /* C0  */  "RESERVEDC0",
    /* C4  */  "RESERVEDC4",
    /* C8  */  "RESERVEDC8",
    /* CC  */  "RESERVEDCC",
    /* D0  */  "RESERVEDD0",
    /* D4  */  "RESERVEDD4",
    /* D8  */  "RESERVEDD8",
    /* DC  */  "RESERVEDDC",
    /* E0  */  "RESERVEDE0",
    /* E4  */  "RESERVEDE4",
    /* E8  */  "RESERVEDE8",
    /* EC  */  "RESERVEDEC",
    /* F0  */  "RESERVEDF0",
    /* F4  */  "RESERVEDF4",
    /* F8  */  "CONREAD",
    /* FC  */  "CONWRITE" };

    if( vax == 0L )
        return VAX_NOVAX;

    addr = vax-> SCBB;
    if( addr == 0L ) {
        printf( "No SCB established (SCBB=0)\n" );
        return VAX_OK;
    }

    count = 0;
    printf( "SYSTEM CONTROL BLOCK\n    SCBB register = %08lX\n", vax-> SCBB );
    saved_mapen = vax-> MAPEN;
    vax-> MAPEN = 0;

    for( n = 0; n < 64; n ++ ) {

        rc = load_memory( vax, addr + ( n * 4 ), ( void * ) &vector, 4 );
        if( rc ) {
           vax-> MAPEN = saved_mapen;
           return rc;
        }

        if( vector || all ) {

			sp = find_label( vax, vector, SYM_LABEL );

			if( count == 0 ) 
				printf( "    Slot  Name            Address  ISP?  Label\n" );
            count++;
            stack = vector & 0x00000003UL;
            vector = vector & 0xFFFFFFFCUL;
            printf( "     %02lX   EXC$%-10s  %08lX %s %s\n", 
              n, vector_name[ n ], vector, stack ? "(ISP)" : "     ",
				  sp == 0L ? "" : sp-> name );
        }
    }

    if( count == 0 )
        printf( "No exception vector entries defined.\n" );

    vax-> MAPEN = saved_mapen;

    return VAX_OK;
}


