//
//  Copyright (C) 1997,1998,1999 Forest Edge Software, All Rights Reserved
//
//  Program:    eVAX, a "Virtual VAX" for Mac OS Computers
//
//  Author:     Tom Cole
//
//  Module:     emul_branch.c
//
//  Purpose:    Emulator handlers for relative branch instructions.
//
//
//  History:    07/28/97    New header format standardization
//
//              01/27/99    Added emul_case() handler


#include "vax.pch"

/*
 *  This handler covers the CASEB, CASEW, and CASEL instructions.
 */

EMULATOR_ENTRY( emul_case )
{


    unsigned long selector, base, limit;
    unsigned long base_displacement, idx;
    long rc;
    short branchword;

    long  * p_long;
    short * p_word;
    char  * p_byte;

//  Based on the operand sizes, fetch the three primary operands.

    switch( opcode-> function ) {

    case 0x8F:   /* CASEB */

        p_byte = ( char * ) get_operand( vax, opcode, 0, OP_RD );
        if( p_byte == 0L )
            return VAX_FAULT;
        selector = ( long ) ( *p_byte );
 
        p_byte = ( char * ) get_operand( vax, opcode, 1, OP_RD );
        if( p_byte == 0L )
            return VAX_FAULT;
        base = ( long ) ( *p_byte );
 
        p_byte = ( char * ) get_operand( vax, opcode, 2, OP_RD );
        if( p_byte == 0L )
            return VAX_FAULT;
        limit = ( long ) ( *p_byte );
 
        break;


    case 0xAF:   /* CASEW */

        p_word = ( short * ) get_operand( vax, opcode, 0, OP_RD );
        if( p_word == 0L )
            return VAX_FAULT;
        selector = ( long ) ( *p_word );
 
        p_word = ( short * ) get_operand( vax, opcode, 1, OP_RD );
        if( p_word == 0L )
            return VAX_FAULT;
        base = ( long ) ( *p_word );
 
        p_word = ( short * ) get_operand( vax, opcode, 2, OP_RD );
        if( p_word == 0L )
            return VAX_FAULT;
        limit = ( long ) ( *p_word );
 
        break;


    case 0xCF:   /* CASEL */

        p_long = ( long * ) get_operand( vax, opcode, 0, OP_RD );
        if( p_long == 0L )
            return VAX_FAULT;
        selector = ( long ) ( *p_long );
 
        p_long = ( long * ) get_operand( vax, opcode, 1, OP_RD );
        if( p_long == 0L )
            return VAX_FAULT;
        base = ( long ) ( *p_long );
 
        p_long = ( long * ) get_operand( vax, opcode, 2, OP_RD );
        if( p_long == 0L )
            return VAX_FAULT;
        limit = ( long ) ( *p_long );
 
        break;

    }

//  At this point, we have the selector, base, and limit values.  The
//  displacement table follows the instruction, so let's find the
//  base displacement location.  

    base_displacement = vax-> PC;

//  Subtract the selector from the base.  This gives us a zero-based
//  index into the list.  Also, this instruction always clears the
//  overflow bit in the PSL.

    idx = selector - base;
    vax-> pslw.v = 0;
    vax-> pslw.n = ( ( long ) idx < ( long ) limit );

//  If the idx is greater than the limit, then we just stride past the
//  table and give it up.

    if( idx > limit ) {

        vax-> pslw.z = 0;
        vax-> pslw.c = 0;

        vax-> PC = base_displacement + 2 + ( limit * 2 );
        return VAX_OK;
    }

//  Otherwise, let's set the PSL bits to reflect the comparison

    vax-> pslw.z = ( idx == limit );
    vax-> pslw.c = ( idx < limit );

//  Calculate the VAX address of the displacement word we want.  Read
//  the displacement and add to the PC.  Note that the base of the
//  displacement starts at the first byte of the table; i.e. the PC
//  is already advanced past the limit operand when we do this read.

    idx = ( idx * 2 ) + base_displacement;

    rc = load_memory( vax, idx, ( void * ) &branchword, 2 );
    if( rc )
        return rc;

    vax-> PC += branchword;
    return VAX_OK;
}


/*
 *  This entry point handles the generic branch instruction families.
 */
        
        
EMULATOR_ENTRY( emul_branch )
{


    long new_pc, rc;
    long * src, data;
    
    //  The operand data already contains the desired address to branch to
    //  should we decide to use it.
    
    new_pc = opcode-> VAXaddr[ 0 ];
    rc = VAX_OK;
    
    switch( opcode-> function ) {
    
    case 0x05:  /* RSB */
    
            /* Read the longword at the stack top into the PC (register 15) */
            
            load_register( vax, 15, vax-> SP, 4 );
            
            /* Now, update the SP accordingly */
            
            vax-> SP = vax-> SP + 4;

            break;
            
    case 0x10:  /* BSBW */
    
            vax->SP -= 4;
            rc = store_memory( vax, vax->SP, ( void * ) &vax->PC, 4 );
            if( !rc )
                vax-> PC = new_pc;
            break;
            
    case 0x11:  /* BRB */
    
            vax-> PC = new_pc;
            break;
            
    case 0x12:  /* BNEQ, BNEQU */
    
            if( vax-> pslw.z == 0 )
                vax-> PC = new_pc;
            break;

    case 0x13:  /* BEQL, BEQLU */
    
            if( vax-> pslw.z == 1 )
                vax-> PC = new_pc;
            break;
            
    case 0x14:  /* BGTR */
    
            if(( vax-> pslw.n | vax-> pslw.z ) == 0 )
                vax-> PC = new_pc;
            break;
    
    case 0x15:  /* BLEQ */
    
            if(( vax-> pslw.n | vax-> pslw.z ) == 1 )
                vax-> PC = new_pc;
            break;
    
    case 0x16:  /* JSB */
    
            vax-> SP -= 4;
            rc = store_memory( vax, vax-> SP, ( void * ) &(vax->PC), 4 );
            if( rc )
                break;
                
            vax-> PC = new_pc;
            break;
            
    case 0x17:  /* JMP */
    
            vax-> PC = opcode-> VAXaddr[ 0 ];
            break;
            
    case 0x18:  /* BGEQ */
    
            if( vax-> pslw.n == 0 )
                vax-> PC = new_pc;
            break;

    case 0x19:  /* BLSS */
    
            if( vax-> pslw.n == 1 )
                vax-> PC = new_pc;
            break;

    case 0x1A:  /* BGTRU */
    
            if(( vax-> pslw.c | vax-> pslw.z ) == 0 )
                vax-> PC = new_pc;
            break;

    case 0x1B:  /* BLEQU */
    
            if(( vax-> pslw.c | vax-> pslw.z ) == 1 )
                vax-> PC = new_pc;
            break;

    case 0x1C:  /* BVC */
    
            if( vax-> pslw.v == 0 )
                vax-> PC = new_pc;
            break;

    case 0x1D:  /* BVS */
    
            if( vax-> pslw.v == 1 )
                vax-> PC = new_pc;
            break;

    case 0x1E:  /* BGEQU, BCC */
    
            if( vax-> pslw.c == 0 )
                vax-> PC = new_pc;
            break;

    case 0x1F:  /* BLSSU, BCS */
    
            if( vax-> pslw.c == 0 )
                vax-> PC = new_pc;
            break;

    case 0x30:  /* BSBW */
    
            vax->SP -= 4;
            rc = store_memory( vax, vax->SP, ( void * ) &vax->PC, 4 );
            if( !rc )
                vax-> PC = new_pc;
            break;
            
    case 0x31:  /* BRW */
    
            vax-> PC = new_pc;
            break;
    
    case 0xE8:  /* BLBS */
    case 0xE9:  /* BLBC */
    
            /* Get the address of the longword we are going to compare */
            
            src = ( long * ) get_operand( vax, opcode, 0, OP_RD );
            if( src == 0L )
                return VAX_FAULT;
            
            /* Read the longword and mask down to the lsb */
            
            data = (*src) & 0x00000001;
            
            /* Depending on which opcode we're doing, test for one or zero */
                        
            if( data == ( opcode-> function == 0xE8 ) ? 1 : 0 )
                vax-> PC = opcode-> VAXaddr[ 1 ];
                
            break;
            
    default:
    
        return set_fault( vax, EXC_PRIV, 0 );
    
    }
    
    return rc;
}


